#include "VirtualRobotReader.h"

#include <optional>
#include <thread>

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/VirtualRobot.h>
#include <VirtualRobot/XML/RobotIO.h>

#include <ArmarXCore/core/PackagePath.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/core/system/cmake/CMakePackageFinder.h>
#include <ArmarXCore/core/time/Clock.h>


namespace armarx::armem::robot_state
{

    VirtualRobotReader::VirtualRobotReader(armem::client::MemoryNameSystem& memoryNameSystem) :
        RobotReader(memoryNameSystem)
    {
    }

    void
    VirtualRobotReader::connect()
    {
        RobotReader::connect();
    }

    // TODO(fabian.reister): register property defs
    void
    VirtualRobotReader::registerPropertyDefinitions(::armarx::PropertyDefinitionsPtr& def)
    {
        RobotReader::registerPropertyDefinitions(def);
    }

    bool
    VirtualRobotReader::synchronizeRobot(VirtualRobot::Robot& robot, const armem::Time& timestamp)
    {
        const auto packages = armarx::CMakePackageFinder::FindAllArmarXSourcePackages();
        const auto package = armarx::ArmarXDataPath::getProject(packages, robot.getFilename());

        const robot::RobotDescription robotDescription{
            .name = robot.getName(), .xml = PackagePath{package, robot.getFilename()}};

        const auto robotState = queryState(robotDescription, timestamp);
        if (not robotState)
        {
            ARMARX_WARNING << "Querying robot state failed for robot `" << robot.getName() << " / "
                           << robot.getType() << "`!";
            return false;
        }

        robot.setJointValues(robotState->jointMap);
        robot.setGlobalPose(robotState->globalPose.matrix());

        return true;
    }

    VirtualRobot::RobotPtr
    VirtualRobotReader::getRobot(const std::string& name,
                                 const armem::Time& timestamp,
                                 const VirtualRobot::RobotIO::RobotDescription& loadMode)
    {
        ARMARX_INFO << "Querying robot description for robot '" << name << "'";
        const auto description = queryDescription(name, timestamp);

        if (not description)
        {
            return nullptr;
        }

        const std::string xmlFilename =
            ArmarXDataPath::resolvePath(description->xml.serialize().path);
        ARMARX_INFO << "Loading (virtual) robot '" << description->name << "' from XML file '"
                    << xmlFilename << "'";

        auto robot = VirtualRobot::RobotIO::loadRobot(xmlFilename, loadMode);
        robot->setName(name);

        synchronizeRobot(*robot, timestamp);

        return robot;
    }


    VirtualRobot::RobotPtr
    VirtualRobotReader::getSynchronizedRobot(
        const std::string& name,
        const armem::Time& timestamp,
        const VirtualRobot::RobotIO::RobotDescription& loadMode,
        const bool blocking)
    {
        while (blocking)
        {
            VirtualRobot::RobotPtr robot = getRobot(name, timestamp, loadMode);
            if (robot and synchronizeRobot(*robot, timestamp))
            {
                return robot;
            }

            ARMARX_INFO << "Retrying to query robot after failure";
            Clock::WaitFor(sleepAfterFailure);
        }

        ARMARX_WARNING << "Failed to get synchronized robot `" << name << "`";
        return nullptr;
    }


} // namespace armarx::armem::robot_state
